/* trashcan.c */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BUFSIZE			1024
#define PATHSIZE		256
#define TRASHFOLDER	".ti3_trash"	


char copy_buffer[BUFSIZE];


/* 'copy' kopiert den Dateiinhalt einer Datei namens 
 * "source". Eine Kopie target wird nur erzeugt, wenn 
 * eine Datei "target" noch nicht existiert.
 * Die Zugriffsrechte werden nicht kopiert, sondern
 * auf "rwx --- ---" gesetzt.
 */
int copy(char *source, char *target)
{
	int source_handle;
	int target_handle;
	source_handle = open(source, O_RDONLY);
	if (source_handle < 0) {
		return -1;
	}
	target_handle = open(target, O_WRONLY | O_CREAT | O_EXCL, S_IWUSR | S_IRUSR | S_IXUSR); 
	if (target_handle <= 0) {
		close(source_handle);
		return -2;
	}
	size_t nread;
	while (1) {
		nread = read(source_handle, copy_buffer, BUFSIZE);
		if (nread <= 0)
			break;
		write(target_handle, copy_buffer, nread);
	}
	close(source_handle);
	close(target_handle);
	return 0;
}


char parse_command(char *command)
{
	return command[1];
}


int setup_trashcan(char *foldername)
{
	return mkdir(foldername, S_IWUSR | S_IRUSR | S_IXUSR);
}


int trash_filepath(char *pathname, char *foldername, char *filename)
{
	strcpy(pathname, foldername);
	strcat(pathname, "/");
	strcat(pathname, filename);
	return 0;
}


int put_file(char *foldername, char *filename)
{
	char pathname[PATHSIZE];
	trash_filepath(pathname, foldername, filename);
	int copy_error = copy(filename, pathname);
	if (copy_error >= 0) {
		if (unlink(filename))
			return -3;
	}
	return copy_error;
}


int get_file(char *foldername, char *filename)
{
	char pathname[PATHSIZE];
	trash_filepath(pathname, foldername, filename);
	int copy_error = copy(pathname, filename);
	if (copy_error >= 0) {
		if (unlink(pathname))
			return -3;
	}
	return copy_error;
}


int remove_file(char *foldername, char *filename)
{
	char pathname[PATHSIZE];
	trash_filepath(pathname, foldername, filename);
	return unlink(pathname);
}


int main(int argc, char *argv[])
{
	if (argc == 1) {
		printf("...not enough arguments!\n");
		return EXIT_FAILURE;
	} else {
		setup_trashcan(TRASHFOLDER);
		char command = parse_command(argv[1]);
		switch (command) {
			case 'p':
				if (argc != 3) {
					printf("...not enough arguments!\n");
					return EXIT_FAILURE;
				}
				int p_error = put_file(TRASHFOLDER, argv[2]);
				switch (p_error) {
					case -1:
						printf("...source file not found!\n");
						break;
					case -2:
						printf("...trash file was not created!\n");
						break;
					case -3:
						printf("...source file could not be removed!\n");
					default:
						break;
				}
				break;
			case 'g':
				if (argc != 3) {
					printf("...not enough arguments!\n");
					return EXIT_FAILURE;
				}
				int g_error = get_file(TRASHFOLDER, argv[2]);
				switch (g_error) {
					case -1:
						printf("...trash file not found!\n");
						break;
					case -2:
						printf("...restore file was not created!\n");
						break;
					case -3:
						printf("...trash file could not be removed!\n");
					default:
						break;
				}
				break;
			case 'r':
				if (argc != 3) {
					printf("...not enough arguments!\n");
					return EXIT_FAILURE;
				}
				int r_error = remove_file(TRASHFOLDER, argv[2]);
				if (r_error)
					printf("...trash file could not be removed!\n");
				break;
			default :
				printf("...unknown command!\n");
				return EXIT_FAILURE;
		}
	}
	return EXIT_SUCCESS;
}
